//   Copyright (c)  2005  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/error.H"
#include "CoCoA/io.H"

#include <iostream>
using std::ostream;
using std::endl;
#include <map>
using std::map;
#include <utility>
using std::make_pair;

namespace CoCoA
{

  // This is simply a forward declaration; the real code is later on
  namespace ErrorLanguage
  {
    static const char* code2message(ERR::code id);
  } // end of namespace ErrorLanguage


  ErrorInfo::ErrorInfo(const std::string& mesg, const std::string& func):
      myCode(ERR::nonstandard),
      myMessage(mesg),
      myContext(func),
      myFile(""),
      myLine(0)
  {}


  ErrorInfo::ErrorInfo(const std::string& mesg, const std::string& func, const char* file, unsigned long line):
      myCode(ERR::nonstandard),
      myMessage(mesg),
      myContext(func),
      myFile(file),
      myLine(line)
  {}


  ErrorInfo::ErrorInfo(ERR::code id, const std::string& func):
      myCode(id),
      myMessage(ErrorLanguage::code2message(id)),
      myContext(func),
      myFile(""),
      myLine(0)
  {}


  ErrorInfo::ErrorInfo(ERR::code id, const std::string& func, const char* file, unsigned long line):
      myCode(id),
      myMessage(ErrorLanguage::code2message(id)),
      myContext(func),
      myFile(file),
      myLine(line)
  {}


  std::ostream& operator<<(std::ostream& out, const ErrorInfo& err)
  {
    return out << "CoCoA::ErrorInfo(\"" << err.what() << "\")";
  }


  void ANNOUNCE(const ErrorInfo& err)
  {
    // Save stream flags to restore them upon exit.
    std::ios::fmtflags OrigFlags = GlobalErrput().flags();
    GlobalErrput().setf(std::ios::dec, std::ios::basefield); // Want line no. in decimal!
    GlobalErrput() << endl
                   << endl
                   << "***************************************************************************" << endl
                   << "****CoCoA ERROR****  Message: " << err.what() << endl;
    if (!err.myContext.empty())
      GlobalErrput() << "****CoCoA ERROR****  Context: " << err.myContext << endl;
    if (!err.myFile.empty())
    {
      GlobalErrput() << "****CoCoA ERROR****  File:    " << err.myFile << endl;
      GlobalErrput() << "****CoCoA ERROR****  Line:    " << err.myLine << endl;
    }
    GlobalErrput() << "***************************************************************************" << endl;
    GlobalErrput() << endl;
    GlobalErrput().setf(OrigFlags);
  }


  void error(const ErrorInfo& err)
  {
#if defined(CoCoA_DEBUG) && CoCoA_DEBUG > 1
    GlobalLogput() << "[Log] Throwing " << err << endl;
#endif
    throw err;
  }


  ///////////////////////////////////////////////////////////////////////////
  // The error messages

  namespace ERR
  {
    const message LANGUAGE("English"),
      nonstandard("Error code corresponding to an error different from the standard CoCoA errors"),
      UNKNOWN("UNKNOWN ERROR CODE PRODUCED -- please inform the CoCoA Team of this"),
      ArgTooBig("Argument to a numerical function too large (value would be too big)"),
      AssertFail("Assertion failed"),
      BadArg("The arg(s) given are unsuitable"),
      BadArraySize("Unsuitable array size (incompatible with other arguments)"),
      BadCheckedCast("Checked cast failed"),
      BadCodomain("Supplied RingHom has wrong codomain"),
      BadColIndex("Column index too large"),
      BadCompose("Attempt to compose maps with differing domain and codomain"),
      BadComptIndex("Index too large accessing component of ModuleElem"),
      BadDegIndex("Index too large accessing degree component"),
      BadIndetIndex("Indeterminate index too large"),
      BadIndetNames("Duplicate indet names or varied number of indices for a single name"),
      BadInducingHom("Inducing hom has wrong domain"),
      BadInducingHom2("Inducing hom has wrong codomain"),
      BadInducingHomKer("Inducing hom has unsuitable kernel"),
      BadMatrixSetEntry("Attempt to set a matrix entry where this is forbidden"),
      BadMatrixSize("Matrix size is wrong for this operation"),
      BadNumBase("Base for number conversion must be between 2 and 36 (incl)"),
      BadNumIndets("Wrong number of indet names when creating a PPMonoid"),
      BadPPMonoid("PPMonoid supplied is not of the right type"),
      BadPolyRingHomImages("Unsuitable indet images for creating PolyRing hom (wrong number or wrong ring)"),
      BadProbability("Probability P must satisfy  0 <= P <= 1"),
      BadPwrZero("Non-positive power of zero"),
      BadQuot("Cannot compute quotient in ring or monoid"),
      BadQuotRing("Attempt to quotient by ideal(1)"),
      BadRing("Used ring does not have desired properties"),
      BadRingHomArg("Argument to RingHom does not belong to domain"),
      BadRingHomArg2("Argument given to a partial RingHom is outside the domain"),
      BadRowIndex("Row index too large"),
      BadSmallFpChar("Characteristic (for small finite field) too large or not prime"),
      BadSymbolHead("Invalid symbolic name head (with illegal characters)"),
      BadSymbolIndex("Symbolic name does not have that many indices"),
      BadSymbolRange("Incompatible symbols given to range"),
      BLASFailed("A BLAS function failed"),
      CanonicalHomFail("Unable to construct canonical homomorphism"),
      ConstMatEntry("Cannot assign to constant matrix entry"),
      DeadMemPool("Attempt to use a MemPool after it has been destroyed"),
      DivByZero("Division by zero or by a zero-divisor"),
      ExpTooBig("Exponent is too large"),
      GlobalManager1("No GlobalManager object has been created"),
      GlobalManager2("Attempt to create more than one GlobalManager object"),
      IdealNotInRing("Incompatible ring and ideal: ideal is not in the given ring"),
      InsuffPrec("RingTwinFloat: insufficient precision (e.g. error growth impedes further computation)"),
      IntDivByNeg("Integer division by negative divisor with non-zero remainder"),
      LapackFailed("A Lapack driver failed"),
      LogZero("Cannot compute log of zero"),
      MemPoolZero("Cannot use MemPool to manage blocks of zero size"),
      MissNumLibs("Numerical libraries not configured in"),
      MixedCoeffRings("Arithmetic operation between polys with different coeff rings"),
      MixedDegrees("Arithmetic operation between incompatible degrees"),
      MixedModules("Elements must be in the same module; mixed module operations are forbidden"),
      MixedPolyIters("Comparison between iterators over different polys"),
      MixedPPMs("Elements must be in the same PPMonoid; mixed PPMonoid operations are forbidden"),
      MixedRings("Elements must be in the same ring; mixed ring operations are forbidden"),
      NegExp("Negative exponent"),
      NoGlobalMgr("GlobalManager must be created before using CoCoALib"),
      NotCommutative("Ring is not commutative"),
      NotDenseUPolyRing("Ring must be a dense univariate polynomial ring"),
      NotEFGModule("Module must be EFGModule, but is not"),
      NotElemFrF("RingElem is not in a fraction field"),
      NotElemGCDDomain("RingElem is not in a GCD domain"),
      NotElemPolyRing("RingElem is not in a polynomial ring"),
      NotElemQuotientRing("RingElem is not in quotient ring"),
      NotElemSparsePolyRing("RingElem is not in a sparse polynomial ring"),
      NotField("Ring must be a field"),
      NotFracField("Ring must be a FractionField"),
      NotFreeModule("Module must be a free module"),
      NotFullRank("Matrix must be full rank"),
      NotGCDDomain("Ring is not a GCD domain"),
      NotGradedModule("Module must be a graded module"),
      NotIndet("Poly should be an indeterminate"),
      NotIntegralDomain("Ring is not an integral domain"),
      NotInvMatrix("Matrix must be invertible"),
      NotNonNegativeGrading("Grading must be non-negative"),
      NotOrdDom("Ring is not an ordered domain"),
      NotPolyRing("Ring must be a polynomial ring"),
      NotPositiveGrading("Grading must be positive"),
      NotQuotientRing("Ring must be a quotient ring"),
      NotRingTwinFloat("Operation valid only over RingTwinFloat"),
      NotRingZ("Ring must be Z, the ring of integers"),
      NotSparsePolyRing("Ring must be a sparse polynomial ring"),
      NotSquareMatrix("Matrix must be square"),
      NotTrueGCDDomain("Ring is not a GCD domain or is a field"),
      NotTermOrdering("Ordering must be a term-ordering (all PPs>1)"),
      NotUnit("Cannot invert non-unit"),
      NYI("NOT YET IMPLEMENTED -- please be patient, we're working on it"),
      PPOrder("PP is not in the right order"),
      PolyIterEnded("Attempt to use an off-the-end iterator"),
      ZeroCoeff("Coefficient must be non-zero"),
      ZeroModulus("A zero modulus was specified for a numerical operation"),
      ZeroPoly("Non-zero polynomial required"),
      SERIOUS("A serious library error has occurred -- please inform the CoCoA Team");

  } // end of namespace ERR


  // These three could be inline, but I don't believe speed is important here.
  // I have defined them outside namespace ERR for improved readability
  // (well, I find ERR::message and ERR::code pleasantly explicit).
  bool ERR::message::operator<(const ERR::code& rhs) const
  {
    return myMessage < rhs.myMessage; // simply compare pointer values
  }

  bool ERR::message::operator==(const ERR::message& rhs) const
  {
    return myMessage == rhs.myMessage;
  }

  bool ERR::message::operator!=(const ERR::message& rhs) const
  {
    return myMessage != rhs.myMessage;
  }


  namespace ErrorLanguage
  {

    typedef map<ERR::code, const char*> MsgTable_t;

    /*const*/ MsgTable_t* GlobalErrorMsgTablePtr = NULL;

    static const char* code2message(ERR::code id)
    {
      if (GlobalErrorMsgTablePtr == 0) return id.myMessage;  // default language
      if (GlobalErrorMsgTablePtr->count(id) > 0)
        return (*GlobalErrorMsgTablePtr)[id];                // return translated message

      // There is no translation of the message associated to id, so use the default message.
      return id.myMessage; // ??? add apology for absence of translated message???
    }


    static void SetLanguage(MsgTable_t& tbl)
    {
      if (tbl.count(ERR::LANGUAGE) == 0) return; // silently refuse to set language if there is no language name
      GlobalErrorMsgTablePtr = &tbl;
    }


    static void set(MsgTable_t& tbl, ERR::code id, const char* const msg)
    {
      if (tbl.insert(std::make_pair(id, msg)).second) return;
      GlobalErrput() << "WARNING: Slot for error `" << id.myMessage << "' already occupied\n"
                     << "WARNING: Failed to register message `" << msg << "'" << endl;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Selecting english is a special cases as it is the default language.

    void english()
    {
      // Setting the pointer should not leak memory as it points to static
      // data when it is not NULL.
      GlobalErrorMsgTablePtr = NULL;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Some Italian error messages are listed here.
    void italian()
    {
      static MsgTable_t MsgTable;
      if (MsgTable.empty())
      {
        set(MsgTable, ERR::LANGUAGE,     "Italian/Italiano");
        set(MsgTable, ERR::nonstandard,  "Codice d'errore che non appartiene all'insieme di errori standard");
        set(MsgTable, ERR::UNKNOWN,      "CODICE D'ERRORE SCONOSCIUTO -- prego di segnalarlo al CoCoA Team");
        set(MsgTable, ERR::SERIOUS,      "Accaduto un errore grave nella libreria CoCoA -- prego di segnalarlo al CoCoA Team");
      }
      SetLanguage(MsgTable);
    }

  } // end of namespace ErrorLanguage

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/error.C,v 1.6 2007/10/30 17:14:06 abbott Exp $
// $Log: error.C,v $
// Revision 1.6  2007/10/30 17:14:06  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.5  2007/10/05 14:35:16  bigatti
// -- added error: NotDenseUPolyRing
//
// Revision 1.4  2007/06/21 21:29:47  abbott
// Changed name of RingFloat into RingTwinFloat.
//
// Revision 1.3  2007/05/21 12:57:28  abbott
// New class for passing machine integers as args; includes some simple
// operations on machine integers (cmp, gcd, IsNegative,...).  Operations
// between ZZ and machine integers modified to use the new class.  Inexact
// integer division (of a ZZ) by a negative value now triggers an error;
// new error for reporting inexact integer division by a negative value.
//
// Revision 1.2  2007/03/28 10:06:13  abbott
// Now gives error when you use RingZ() or RingQ() without creating GlobalManager.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.12  2007/03/08 18:22:28  cocoa
// Just whitespace cleaning.
//
// Revision 1.11  2007/03/08 11:07:12  cocoa
// Made pseudo ctors for polynomial rings more uniform.  This allowed me to
// remove an include of CoCoA/symbol.H  from the RingDistrM*.H files, but then
// I had to put the include in several .C files.
//
// Revision 1.10  2007/03/05 21:25:57  cocoa
// Forgot to check these in a few minutes ago.
//
// Revision 1.9  2007/03/05 16:17:11  bigatti
// -- clenup for numerical code (and 3 new error codes)
//
// Revision 1.8  2007/03/03 14:07:23  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.7  2007/03/02 10:47:53  cocoa
// First stage of RingZ modifications -- tests do not compile currently, Anna will fix this.
//
// Revision 1.6  2007/02/12 17:40:19  bigatti
// -- added MissNumLibs
//
// Revision 1.5  2007/01/13 14:14:34  cocoa
// Overhaul of RingHom code: it nows uses SmartPtrIRC, and printing is more logical.
// Have not yet updated the documentation.
//
// Revision 1.4  2006/12/21 13:48:32  cocoa
// Made all increment/decrement calls prefix (except where the must be postfix).
//
// Revision 1.3  2006/08/07 21:23:25  cocoa
// Removed almost all publicly visible references to SmallExponent_t;
// changed to long in all PPMonoid functions and SparsePolyRing functions.
// DivMask remains to sorted out.
//
// Revision 1.2  2006/07/17 19:16:53  cocoa
// New, better errors :-)
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.9  2006/04/14 13:51:40  cocoa
// Added a class for generating random bits (RandomBitStream in the file random.H).
//
// Revision 1.8  2006/04/07 16:44:52  cocoa
// Considerably updated MemPool design -- it works, and I'm about to test
// its efficiency against the old one.
//
// Revision 1.7  2006/04/04 13:01:02  cocoa
// -- added: NotNonNegativeGrading, NotPositiveGrading
//
// Revision 1.6  2006/03/27 12:21:25  cocoa
// Minor silly changes to reduce number of complaints from some compiler or other.
//
// Revision 1.5  2006/03/12 21:28:33  cocoa
// Major check in after many changes
//
// Revision 1.4  2005/12/31 12:22:17  cocoa
// Several minor tweaks to silence the Microsoft compiler:
//  - added some missing #includes and using directives
//  - moved some function defns into the right namespace
//  - etc.
//
// Revision 1.3  2005/11/24 16:09:38  cocoa
// -- added operator[] for ModuleElem
//
// Revision 1.2  2005/10/26 14:34:51  cocoa
// Changed ANNOUNCE: it now sets sane ostream flags for itself, and resets
// the original flags before exiting.  No more line numbers in hexadecimal!
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.8  2005/09/28 11:50:34  cocoa
// -- new code for graded modules
//
// Revision 1.7  2005/08/08 16:36:32  cocoa
// Just checking in before going on holiday.
// Don't really recall what changes have been made.
// Added IsIndet function for RingElem, PPMonoidElem,
// and a member function of OrdvArith.
// Improved the way failed assertions are handled.
//
// Revision 1.6  2005/07/19 15:30:20  cocoa
// A first attempt at iterators over sparse polynomials.
// Main additions are to SparsePolyRing, DistrMPoly*.
// Some consequential changes to PPMonoid*.
//
// Revision 1.5  2005/07/15 16:34:33  cocoa
// Added iterators for sparse polynomials.
// The code compiles (and the old tests still run).
// It'd Friday evening -- I'm going home before
// getting any ideas about making the iterator code run.
//
// Revision 1.4  2005/07/08 15:09:28  cocoa
// Added new symbol class (to represent names of indets).
// Integrated the new class into concrete polynomial rings
// and PPMonoid -- many consequential changes.
// Change ctors for the "inline" sparse poly rings: they no
// longer expect a PPMonoid, but build their own instead
// (has to be a PPMonoidOv).
//
// Revision 1.3  2005/07/01 16:08:15  cocoa
// Friday check-in.  Major change to structure under PolyRing:
// now SparsePolyRing and DUPolyRing are separated (in preparation
// for implementing iterators).
//
// A number of other relatively minor changes had to be chased through
// (e.g. IndetPower).
//
// Revision 1.2  2005/06/27 14:55:24  cocoa
// Cleaned up some more PPMonoid code.
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.8  2005/04/29 15:42:02  cocoa
// Improved documentation for GMPAllocator.
// Added example program for GMPAllocator.
// Added example program for simple ops on polynomials.
// Added two new ctors for (principal) ideals (from long, and from ZZ).
// Added (crude) printing for PPMonoids.
// Updated library.H (#included GMPAllocator.H).
//
// Revision 1.7  2005/04/27 16:14:56  cocoa
// Cleaned up example programs -- added "free use" permit.
// Changed a couple of ErrorInfo object names, and added
// ERR::NotTrueGCDDomain.
//
// Revision 1.6  2005/04/20 15:40:47  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.5  2005/04/19 14:06:03  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.4  2005/03/11 16:44:18  cocoa
// New abstract class structure for matrices.
// New types of special matrix.
//
// Revision 1.3  2005/03/02 18:46:41  cocoa
// Added new types ConstRefMatrix, and RefMatrix following along
// the lines of ConstRefRingElem and RefRingElem.  The semantics
// should be a bit clearer now.
//
// Revision 1.2  2005/02/22 12:50:40  cocoa
// -- added: NotFullRank, NotIntDom, NotTermOrdering
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.17  2004/11/25 16:14:21  cocoa
// (1) Fixed definition of specialization of std::swap template function
//     so that it compiles with gcc 3.4.3
// (2) Implemented monomial function for polynomial rings.
// (3) Added one(PPM) and PPM->myOne() functions.
//
// Revision 1.16  2004/11/19 17:43:50  cocoa
// -- added matrix related errors
//
// Revision 1.15  2004/11/19 16:15:51  cocoa
// (a) Removed unused error message about degree of zero;
//     replaced it by a more general message about needing a
//     non-zero polynomial (for various ops such as LC, LPP).
// (b) Added some missing arg checking in LC, LPP and deg
//     (for elems of a PolyRing).
// (c) Updated some commented out code in GPair and GPoly.
//
// Revision 1.14  2004/11/19 15:14:09  cocoa
// (a) Added new check to MemPool so that it can signal an
//     error if one tries to use a MemPool after it has been
//     destroyed.
// (b) Improved makefile in TEST/ so that it checks output,
//     and prints useful messages if the test fails.
// (c) Tidied ring.txt a bit (still more to do).
//
// Revision 1.13  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.12  2004/11/11 11:56:09  cocoa
// (1) Tidied makefiles, and introduced common.mki
// (2) Improved several tests, and cleaned them so that they
//     handle sanely any otherwise unhandled exceptions.
//
// Revision 1.11  2004/11/09 15:53:57  cocoa
// -- added BadQuotRing("Attempt to quotient by ideal(1)")
//
// Revision 1.10  2004/11/05 15:37:56  cocoa
// Added a couple of new errors.
//
// Revision 1.9  2004/11/03 18:01:55  cocoa
// -- clarified behaviour of error()
// -- added ANNOUNCE function
// -- minor tidying
//
// Revision 1.7  2004/07/27 16:03:38  cocoa
// Added IsCommutative test and IamCommutative member function
// to all rings.  Tidied geobuckets a little.
//
// Revision 1.6  2004/07/14 16:40:42  cocoa
// Separated RingFpLog from its implementation which now resides in
// a new class: SmallFpLogImpl.  This is analogous to the change made
// to RingFp yesterday.
//
// Some tidying and other sundry minor changes.
//
// Revision 1.5  2004/07/13 16:32:26  cocoa
// First stage of major revamp of ring elements.
// Implementation of RingFp has been split into "ring interface"
// and "algorithms plus data structures".
//
// Revision 1.4  2004/06/29 17:10:22  cocoa
// Partially tidied use of "protected" and "private" in various
// base classes.  Checking in at the end of the day -- it works,
// and I wouldn't want it to be lost next time point's disk
// misbehaves.
//
// Revision 1.3  2004/05/27 16:14:02  cocoa
// Minor revision for new coding conventions.
//
// Revision 1.2  2004/05/24 15:52:13  cocoa
// Major update:
//   new error mechanism
//   many fixes
//   RingHoms almost work now
//   RingFloat much improved
//
// Revision 1.1.1.1  2003/09/24 12:55:43  cocoa
// Imported files
//
// Revision 1.4  2003/06/23 16:15:21  abbott
// Minor cleaning prior to public release.
//
// Revision 1.3  2002/11/13 15:41:57  abbott
// The definition of CoCoA::error has been improved.  It now preceeds
// each output line with "CoCoA ERROR", it outputs on std::cerr
// (instead of std::cout), and it will print out the location information
// whenever present.  The function finally throws the exception whereas
// previously it just called exit -- throwing should be cleaner since
// stack unwinding occurs.
//
// Revision 1.2  2001/12/07 15:56:43  abbott
// error function now takes an exception as an argument, instead of a string.
//
// Revision 1.1  2001/11/26 19:13:33  abbott
// Initial revision
//
